import java.util.*;

class PPMC extends MCQ {
    Vertex[][] V;   
    ArrayList<Integer> DeleteList;
    ArrayList[] degreeClass;

    PPMC (int n,boolean[][] A,int[] degree,int style) {
	super(n,A,degree,style);
    }


    void search(){
	cpuTime       = System.currentTimeMillis();
	nodes         = 0;
	colourClass   = new ArrayList[n];
	degreeClass   = new ArrayList[n];
	SparseSet C   = new SparseSet(n);
	SparseSet P   = new SparseSet(n);
	V             = new Vertex[n][n];
	DeleteList    = new ArrayList<Integer>(n);
	for (int i=0;i<n;i++) colourClass[i] = new ArrayList<Vertex>(n);
	for (int i=0;i<n;i++) degreeClass[i] = new ArrayList<Vertex>(n);
	for (int i=0;i<n;i++) P.add(i);
	for (int i=0;i<n;i++) 
	    for (int j=0;j<n;j++) V[i][j] = new Vertex(j,degree[j]);
	expand(0,C,P);
    }

    void expand(int depth,SparseSet C,SparseSet P){
	if (timeLimit > 0 && System.currentTimeMillis() - cpuTime >= timeLimit) return;
	nodes++;
	int m = P.size();
	Vertex[] W = new Vertex[m];
	int k = 0;
	for (int i=0;i<m;i++) W[k++] = V[depth][P.get(i)];
	if (style == 1) sequentialColour(W);
	if (style == 2) brelazColour(W);
	for (int i=m-1;i>=0;i--){
	    Vertex v = W[i];
	    if (C.size() + v.colour <= maxSize) return;
	    for (int j=0;j<=i;j++) V[depth+1][P.get(j)].degree = V[depth][P.get(j)].degree;
	    P.push(); C.add(v.index);
	    select(depth+1,v.index,P);
	    if (P.isEmpty() && C.size() > maxSize) saveSolution(C);
	    if (!P.isEmpty()) expand(depth+1,C,P);
	    P.pop(); C.remove(v.index);
	    reject(depth,v.index,P);
	}
    }

    void select(int depth,int v,SparseSet P){
	DeleteList.clear();
	for (int i=0;i<P.size();i++){
	    int u = P.get(i);
	    if (!A[u][v]) DeleteList.add(u);
	} 
	remove(depth,P);
    }

    void reject(int depth,int v,SparseSet P){
	DeleteList.clear();
	DeleteList.add(v);
	remove(depth,P);
    }

    void remove(int depth,SparseSet P){
	for (int i=0;i<DeleteList.size();i++) P.remove(DeleteList.get(i));
	for (int i=0;i<DeleteList.size();i++){
	    int u = DeleteList.get(i);
	    for (int j=0;j<P.size();j++){
		int w = P.get(j);
		    if (A[u][w]) V[depth][w].degree--;
	    }
	}
    }

    void sequentialColour(Vertex[] W){
	int colours = 0;
	int m = W.length;
	for (int i=0;i<m;i++) colourClass[i].clear();
	for (int i=0;i<m;i++) degreeClass[i].clear();	
	for (int i=m-1;i>=0;i--) degreeClass[W[i].degree].add(W[i]);
	int count = 0;
	for (int i=0;i<m;i++)
	    for (int j=0;j<degreeClass[i].size();j++)
		W[count++] = (Vertex)(degreeClass[i].get(j));
	for (int i=m-1;i>=0;i--){
	    Vertex v = W[i];
	    int k = 0;
	    while (conflicts(v,colourClass[k])) k++;
	    colourClass[k].add(v);
	    v.colour = k+1;
	    colours = Math.max(colours,v.colour);
	}
	count = 0;
	for (int k=0;k<colours;k++)
	    for (int j=0;j<colourClass[k].size();j++){
		Vertex v = (Vertex)(colourClass[k].get(j));
		W[count++] = v;
	    }
    }

    boolean conflicts(Vertex v,ArrayList<Vertex> colourClass){
	for (int i=0;i<colourClass.size();i++){
	    Vertex w = colourClass.get(i);
	    if (A[v.index][w.index]) return true;
	}
	return false;
    }

    void brelazColour(Vertex[] W){
	int colours = 0;
	int m = W.length;
	for (int i=0;i<m;i++) colourClass[i].clear();
	ArrayList<Vertex> L = new ArrayList<Vertex>(m);
	for (Vertex v : W){v.init(m); L.add(v);}
	while (!L.isEmpty()){
	    Vertex v = selectVertex(L);
	    L.remove(v);
	    colourVertex(v,L);
	    colourClass[v.colour - 1].add(v);
	    colours = Math.max(colours,v.colour);
	}
	int i = 0;
	for (int k=0;k<colours;k++)
	    for (int j=0;j<colourClass[k].size();j++){
		Vertex v = (Vertex)(colourClass[k].get(j));
		W[i++] = v;
	    }
    }

    Vertex selectVertex(ArrayList<Vertex> L){
	int maxSaturation = -1,
	    maxDegree     = -1;
        Vertex selected   = null;
	for (int i=0;i<L.size();i++){
	    Vertex v = L.get(i);
	    if (v.saturation > maxSaturation || 
		v.saturation == maxSaturation && v.degree > maxDegree){
		maxSaturation = v.saturation;
		maxDegree     = v.degree;
		selected      = v;
	    }
	}
	return selected;
    }
    
    void colourVertex(Vertex v, ArrayList<Vertex> L){
	int k = v.colour();
	v.colour = k+1;
	for (int i=0;i<L.size();i++){
	    Vertex w = L.get(i);
	    if (A[v.index][w.index]) w.removeColour(k);
	}
    }
    
    void saveSolution(SparseSet C){
	Arrays.fill(solution,0);
	for (int i=0;i<C.size();i++) solution[C.get(i)] = 1;
	maxSize = C.size();
    }
}
